"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const enumerable_1 = require("./enumerable");
const equality_comparer_1 = require("./equality-comparer");
const exceptions_1 = require("./exceptions");
const key_value_pair_1 = require("./key-value-pair");
const utils = require("./utils");
class Dictionary {
    constructor(dictionary, comparer) {
        this.freeList = -1;
        this.freeCount = 0;
        this.version = 0;
        this.count = 0;
        this.comparer = comparer || equality_comparer_1.EqualityComparer.Default();
        if (dictionary) {
            if (typeof dictionary === 'number') {
                let capacity = dictionary;
                if (capacity < 0) {
                    throw new exceptions_1.ArgumentOutOfRangeException('capacity', capacity, 'capacity 小于 0');
                }
                this.Initialize(capacity);
            }
            else {
                for (let item of dictionary) {
                    if (item instanceof key_value_pair_1.KeyValuePair) {
                        this.Set(item.Key, item.Value);
                    }
                    else {
                        this.Set(item[0], item[1]);
                    }
                }
            }
        }
    }
    Initialize(capacity) {
        let size = HashHelpers.GetPrime(capacity);
        let buckets = new Array(size);
        let entries = new Array(size);
        for (let i = 0; i < size; i++) {
            buckets[i] = -1;
            entries[i] = new Entry();
        }
        this.buckets = buckets;
        this.entries = entries;
        this.freeList = -1;
        return size;
    }
    Add(key, value) {
        this.TryInsert(key, value, InsertionBehavior.ThrowOnExisting);
    }
    Set(key, value) {
        this.TryInsert(key, value, InsertionBehavior.OverwriteExisting);
    }
    Get(key) {
        let i = this.FindEntry(key);
        if (i >= 0) {
            let entries = this.entries;
            return entries[i].value;
        }
        throw new exceptions_1.KeyNotFoundException();
    }
    ContainsKey(key) {
        return this.FindEntry(key) >= 0;
    }
    Clear() {
        if (this.count > 0) {
            let buckets = this.buckets;
            let entries = this.entries;
            for (let i = 0; i < buckets.length; i++) {
                buckets[i] = -1;
            }
            for (let i = 0; i < entries.length; i++) {
                entries[i] = new Entry();
            }
            this.freeList = -1;
            this.count = 0;
            this.freeCount = 0;
            this.version++;
        }
    }
    Remove(key) {
        utils.throws.ThrowIfNull('key', key);
        let buckets = this.buckets;
        let entries = this.entries;
        let comparer = this.comparer;
        if (buckets) {
            let hashCode = comparer.GetHashCode(key) & 0x7fffffff;
            let bucket = hashCode % buckets.length;
            let last = -1;
            for (let i = buckets[bucket]; i >= 0; last = i, i = entries[i].next) {
                if (entries[i].hashCode === hashCode &&
                    comparer.Equals(entries[i].key, key)) {
                    if (last < 0) {
                        buckets[bucket] = entries[i].next;
                    }
                    else {
                        entries[last].next = entries[i].next;
                    }
                    entries[i].hashCode = -1;
                    entries[i].next = this.freeList;
                    entries[i].key = undefined;
                    entries[i].value = undefined;
                    this.freeList = i;
                    this.freeCount++;
                    this.version++;
                    return true;
                }
            }
        }
        return false;
    }
    TryGetValue(key) {
        let i = this.FindEntry(key);
        if (i >= 0) {
            let entries = this.entries;
            let value = entries[i].value;
            return [true, value];
        }
        return [false, undefined];
    }
    FindEntry(key) {
        utils.throws.ThrowIfNull('key', key);
        let buckets = this.buckets;
        let entries = this.entries;
        let comparer = this.comparer;
        if (buckets) {
            let hashCode = comparer.GetHashCode(key) & 0x7fffffff;
            for (let i = buckets[hashCode % buckets.length]; i >= 0; i = entries[i].next) {
                if (entries[i].hashCode === hashCode &&
                    comparer.Equals(entries[i].key, key)) {
                    return i;
                }
            }
        }
        return -1;
    }
    TryInsert(key, value, behavior) {
        utils.throws.ThrowIfNull('key', key);
        if (!this.buckets) {
            this.Initialize(0);
        }
        let buckets = this.buckets;
        let entries = this.entries;
        let comparer = this.comparer;
        let hashCode = comparer.GetHashCode(key) & 0x7fffffff;
        let targetBucket = hashCode % this.buckets.length;
        for (let i = buckets[targetBucket]; i >= 0; i = entries[i].next) {
            if (entries[i].hashCode === hashCode &&
                comparer.Equals(entries[i].key, key)) {
                if (behavior === InsertionBehavior.ThrowOnExisting) {
                    throw new exceptions_1.ArgumentException('key', '已经存在相同的Key');
                }
                entries[i].value = value;
                this.version++;
                return true;
            }
        }
        let index = 0;
        if (this.freeCount > 0) {
            index = this.freeList;
            this.freeList = entries[index].next;
            this.freeCount--;
        }
        else {
            if (this.count === entries.length) {
                this.Resize();
                entries = this.entries;
                buckets = this.buckets;
                targetBucket = hashCode % buckets.length;
            }
            index = this.count;
            this.count++;
        }
        entries[index].hashCode = hashCode;
        entries[index].next = buckets[targetBucket];
        entries[index].key = key;
        entries[index].value = value;
        buckets[targetBucket] = index;
        this.version++;
    }
    Resize() {
        let newSize = HashHelpers.GetPrime(this.count * 2);
        let newBuckets = new Array();
        for (let i = 0; i < newSize; i++) {
            newBuckets[i] = -1;
        }
        let newEntries = new Array();
        let entries = this.entries;
        for (let i = 0; i < newSize; i++) {
            if (i < this.count) {
                newEntries[i] = entries[i];
            }
            else {
                newEntries[i] = new Entry();
            }
        }
        for (let i = 0; i < this.count; i++) {
            let bucket = newEntries[i].hashCode % newSize;
            newEntries[i].next = newBuckets[bucket];
            newBuckets[bucket] = i;
        }
        this.buckets = newBuckets;
        this.entries = newEntries;
    }
    *[Symbol.iterator]() {
        if (this.entries) {
            for (let item of this.entries) {
                if (item.key !== undefined) {
                    yield new key_value_pair_1.KeyValuePair(item.key, item.value);
                }
            }
        }
    }
    All(predicate) {
        return enumerable_1.Enumerable.All(this, predicate);
    }
    Any(predicate) {
        return enumerable_1.Enumerable.Any(this, predicate);
    }
    AsEnumerable() {
        return enumerable_1.Enumerable.AsEnumerable(this);
    }
    Average(selector) {
        return enumerable_1.Enumerable.Average(this, selector);
    }
    Concat(second) {
        return enumerable_1.Enumerable.Concat(this, second);
    }
    Contains(value, comparer) {
        return enumerable_1.Enumerable.Contains(this, value, comparer);
    }
    Count(predicate) {
        if (predicate) {
            return enumerable_1.Enumerable.Count(this, predicate);
        }
        else {
            return this.count - this.freeCount;
        }
    }
    DefaultIfEmpty(defaultValue) {
        return enumerable_1.Enumerable.DefaultIfEmpty(this, defaultValue);
    }
    Distinct(comparer) {
        return enumerable_1.Enumerable.Distinct(this, comparer);
    }
    ElementAt(index) {
        return enumerable_1.Enumerable.ElementAt(this, index);
    }
    ElementAtOrDefault(defaultValue, index) {
        return enumerable_1.Enumerable.ElementAtOrDefault(this, defaultValue, index);
    }
    Except(second, comparer) {
        return enumerable_1.Enumerable.Except(this, second, comparer);
    }
    First(predicate) {
        return enumerable_1.Enumerable.First(this, predicate);
    }
    FirstOrDefault(defaultValue, predicate) {
        return enumerable_1.Enumerable.FirstOrDefault(this, defaultValue, predicate);
    }
    GroupBy(keySelector, elementSelector, comparer) {
        return enumerable_1.Enumerable.GroupBy(this, keySelector, elementSelector, comparer);
    }
    GroupJoin(inner, outerKeySelector, innerKeySelector, resultSelector, comparer) {
        return enumerable_1.Enumerable.GroupJoin(this, inner, outerKeySelector, innerKeySelector, resultSelector, comparer);
    }
    Intersect(second, comparer) {
        return enumerable_1.Enumerable.Intersect(this, second, comparer);
    }
    Join(inner, outerKeySelector, innerKeySelector, resultSelector, comparer) {
        return enumerable_1.Enumerable.Join(this, inner, outerKeySelector, innerKeySelector, resultSelector, comparer);
    }
    Last(predicate) {
        return enumerable_1.Enumerable.Last(this, predicate);
    }
    LastOrDefault(defaultValue, predicate) {
        return enumerable_1.Enumerable.LastOrDefault(this, defaultValue, predicate);
    }
    Max(selector) {
        return enumerable_1.Enumerable.Max(this, selector);
    }
    Min(selector) {
        return enumerable_1.Enumerable.Min(this, selector);
    }
    Reverse() {
        return enumerable_1.Enumerable.Reverse(this);
    }
    Select(selector) {
        return enumerable_1.Enumerable.Select(this, selector);
    }
    SelectMany(collectionSelector, resultSelector) {
        return enumerable_1.Enumerable.SelectMany(this, collectionSelector, resultSelector);
    }
    SequenceEqual(second, comparer) {
        return enumerable_1.Enumerable.SequenceEqual(this, second, comparer);
    }
    Single(predicate) {
        return enumerable_1.Enumerable.Single(this, predicate);
    }
    SingleOrDefault(defaultValue, predicate) {
        return enumerable_1.Enumerable.SingleOrDefault(this, defaultValue, predicate);
    }
    Skip(count) {
        return enumerable_1.Enumerable.Skip(this, count);
    }
    SkipWhile(predicate) {
        return enumerable_1.Enumerable.SkipWhile(this, predicate);
    }
    Sum(selector) {
        return enumerable_1.Enumerable.Sum(this, selector);
    }
    Take(count) {
        return enumerable_1.Enumerable.Take(this, count);
    }
    TakeWhile(predicate) {
        return enumerable_1.Enumerable.TakeWhile(this, predicate);
    }
    ToArray() {
        return enumerable_1.Enumerable.ToArray(this);
    }
    ToDictionary(keySelector, elementSelector, comparer) {
        return enumerable_1.Enumerable.ToDictionary(this, keySelector, elementSelector, comparer);
    }
    ToList() {
        return enumerable_1.Enumerable.ToList(this);
    }
    Union(second, comparer) {
        return enumerable_1.Enumerable.Union(this, second, comparer);
    }
    Where(predicate) {
        return enumerable_1.Enumerable.Where(this, predicate);
    }
}
exports.Dictionary = Dictionary;
class Entry {
    constructor() {
        this.hashCode = 0;
        this.next = -1;
        this.key = undefined;
        this.value = undefined;
    }
}
var InsertionBehavior;
(function (InsertionBehavior) {
    InsertionBehavior[InsertionBehavior["None"] = 0] = "None";
    InsertionBehavior[InsertionBehavior["OverwriteExisting"] = 1] = "OverwriteExisting";
    InsertionBehavior[InsertionBehavior["ThrowOnExisting"] = 2] = "ThrowOnExisting";
})(InsertionBehavior || (InsertionBehavior = {}));
class HashHelpers {
    static IsPrime(candidate) {
        if ((candidate & 1) !== 0) {
            let limit = parseInt(Math.sqrt(candidate).toFixed(0), 10);
            for (let divisor = 3; divisor <= limit; divisor += 2) {
                if (candidate % divisor === 0) {
                    return false;
                }
            }
            return true;
        }
        return candidate === 2;
    }
    static GetPrime(min) {
        for (let i = 0; i < PRIMES.length; i++) {
            let prime = PRIMES[i];
            if (prime >= min) {
                return prime;
            }
        }
        for (let i = min | 1; i < 2147483647; i += 2) {
            if (HashHelpers.IsPrime(i) && (i - 1) % HashHelpers.HashPrime !== 0) {
                return i;
            }
        }
        return min;
    }
}
HashHelpers.HashPrime = 101;
const PRIMES = [
    3,
    7,
    11,
    17,
    23,
    29,
    37,
    47,
    59,
    71,
    89,
    107,
    131,
    163,
    197,
    239,
    293,
    353,
    431,
    521,
    631,
    761,
    919,
    1103,
    1327,
    1597,
    1931,
    2333,
    2801,
    3371,
    4049,
    4861,
    5839,
    7013,
    8419,
    10103,
    12143,
    14591,
    17519,
    21023,
    25229,
    30293,
    36353,
    43627,
    52361,
    62851,
    75431,
    90523,
    108631,
    130363,
    156437,
    187751,
    225307,
    270371,
    324449,
    389357,
    467237,
    560689,
    672827,
    807403,
    968897,
    1162687,
    1395263,
    1674319,
    2009191,
    2411033,
    2893249,
    3471899,
    4166287,
    4999559,
    5999471,
    7199369
];
